from interrogatio.core.exceptions import AlreadyRegisteredError
from interrogatio.validators.base import Validator


class ValidatorsRegistry(dict):
    def register(self, alias, clazz):
        if alias in self:
            raise AlreadyRegisteredError(
                'validator `{}` already exists.'.format(alias))
        self[alias] = clazz

    def get_registered(self):
        return list(self.keys())

    def get_instance(self, v):
        clazz = self[v['name']]
        if 'args' in v:
            return clazz(**v['args'])
        return clazz()


_registry = ValidatorsRegistry()


def get_registry():
    return _registry


def register(name):
    if name in get_registry().get_registered():
        raise AlreadyRegisteredError(
            f'The validator `{name}` is already registered.',
        )

    def _wrapper(cls):
        if not issubclass(cls, Validator):
            raise ValueError(
                'The provided class must be a subclass of Validator.',
            )

        get_registry().register(name, cls)

        return cls
    return _wrapper


def get_registered():
    """
    Returns a list of registered Validators.

    :return: list of aliases of the registered Validators.
    :rtype: list
    """
    return get_registry().get_registered()


def get_instance(v):
    """
    Returns an instance of a concrete subclass of Validator.

    :param v: a validator definition object.
    :type v: dict

    :return: an instance of the corresponding Validator.
    :rtype: Validator subclass
    """
    return get_registry().get_instance(v)



from pkg_resources import DistributionNotFound, get_distribution

from interrogatio.core.dialog import dialogus
from interrogatio.core.prompt import interrogatio

__all__ = ('dialogus', 'interrogatio')


try:
    __version__ = get_distribution('interrogatio').version
except DistributionNotFound:  # pragma: no cover
    __version__ = '0.0.0'


def get_version():
    return __version__



from interrogatio.core.exceptions import AlreadyRegisteredError
from interrogatio.handlers.base import QHandler


class QHandlersRegistry(dict):
    def register(self, alias, clazz):
        if alias in self:
            raise AlreadyRegisteredError(f'alias `{alias}` already exists.')
        self[alias] = clazz

    def get_registered(self):
        return list(self.keys())

    def get_instance(self, question):
        qtype = question['type']
        clazz = self[qtype]
        return clazz(question)


_registry = QHandlersRegistry()


def get_registry():
    return _registry


def get_instance(question):
    """
    Returns an instance of a concrete subclass of QHandler initialized with
    a question. The subclass of QHandler is identified by the ``type`` field
    of the question.

    :param question: a question to handle.
    :type question: dict

    :return: an instance of the corresponding QHandler
    :rtype: QHandler subclass
    """
    return get_registry().get_instance(question)


def get_registered():
    """
    Returns a list of registered QHandlers.

    :return: list of aliases of the registered QHandlers.
    :rtype: list
    """
    return get_registry().get_registered()


def register(name):
    if name in get_registry().get_registered():
        raise AlreadyRegisteredError(
            f'The handler `{name}` is already registered.',
        )

    def _wrapper(cls):
        if not issubclass(cls, QHandler):
            raise ValueError(
                'The provided class must be a subclass of QHandler.',
            )

        get_registry().register(name, cls)

        return cls
    return _wrapper



from interrogatio.validators.base import Validator  # noqa
from interrogatio.validators.builtins import *  # noqa
from interrogatio.validators.registry import (  # noqa
    get_instance,
    get_registered,
    register,
)



from abc import ABCMeta, abstractmethod

from prompt_toolkit.application.current import get_app
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from interrogatio.core.exceptions import ValidationError

__all__ = [
    'QHandler',
]


class QHandler(metaclass=ABCMeta):
    """
    ABC for question handlers.
    Each question handler must subclass this class.
    """

    def __init__(self, question):
        self._question = question
        self._widget = None
        self._errors = []

    @property
    def errors(self):
        """
        This property holds a list of validation error messages.
        The list is filled after a call to the is_valid method.

        :return: list of error messages.
        :rtype: list
        """
        return self._errors

    @abstractmethod
    def get_layout(self):
        """
        Returns the UI layout of the question. It must returns a
        python-prompt-toolkit layout container.

        Subclasses must implement this method.

        :return: the UI layout of the question.
        :rtype: :class:`~prompt_toolkit.layout.Layout`
        """
        raise NotImplementedError(
            'Subclass must implements `get_layout` method.',
        )

    @abstractmethod
    def get_value(self):
        """
        Returns the ``value`` part of the answer.

        Subclasses must implement this method.

        :return: the ``value`` part of the answer.
        :rtype: str
        """
        raise NotImplementedError(
            'Subclass must implements `get_value` method.',
        )

    def get_formatted_value(self):
        return self.get_value()

    @abstractmethod
    def get_widget_init_kwargs(self):
        """
        Returns the keyword arguments needed to instantiate the widget.

        Subclasses must implement this method.

        :return: a dictionary containing the keyword arguments.
        :rtype: dict
        """
        raise NotImplementedError(
            'Subclass must implements `get_widget_init_kwargs` method.',
        )

    def get_init_extra_args(self):
        """
        Returns extra arguments needed to instantiate a QHandler.
        The extra arguments are represented as a dictionary within the
        question object under the key ``extra_args``.

        :return: a dictionary containing the initialization extra arguments.
        :rtype: dict
        """
        return self._question.get('extra_args', dict())

    @abstractmethod
    def get_widget_class(self):
        """
        Returns the widget class for this QHandler.

        :return: a widget class.
        :rtype: class
        """
        raise NotImplementedError(
            'Subclass must implements `get_widget_class` method.',
        )

    def get_widget(self):
        """
        Returns the widget instance for this QHandler.
        If the widget has not been already created, this method creates it
        before returns.
        """
        if not self._widget:
            clazz = self.get_widget_class()
            self._widget = clazz(**self.get_widget_init_kwargs())
        return self._widget

    def get_keybindings(self):
        """
        Returns a KeyBindings object to add custom keybindings to this
        QHandler.
        """
        bindings = KeyBindings()

        @bindings.add(Keys.ControlC)
        def _ctrl_c(event):
            get_app().exit(result=False)

        @bindings.add(Keys.Enter)
        def _enter(event):
            get_app().exit(result=True)

        return bindings

    def get_answer(self):
        """
        Returns dictionary with the question variable as key and the answer
        as the value.
        """
        return {self._question['name']: self.get_value()}

    def get_variable_name(self):
        """
        Returns the name of the variable of this question.
        """
        return self._question['name']

    def is_valid(self):
        """
        Apply any speficied validator to the answer and return True if the
        input is valid otherwise False.
        If the answer isn't valid, it also set the errors property to a list
        of error messages.
        """
        validators = self._question.get('validators', [])
        self._errors = []
        for validator in validators:
            try:
                validator.validate(self.get_value())
            except ValidationError as ve:
                self._errors.append(str(ve))
        return not self._errors



from abc import ABCMeta, abstractmethod


__all__ = [
    'Validator',
]


class Validator(metaclass=ABCMeta):
    """
    Abstract class for validators.
    """
    def __init__(self, message='invalid input'):
        self.message = message

    @abstractmethod
    def validate(self, value):
        """
        Abstract method.
        Subclasses must implement this method with the validation logic.

        :param value: the value to validate.
        :type value: str

        :raises:
            ValidationError: if the provided input is invalid.
        """



from prompt_toolkit.application import get_app
from prompt_toolkit.filters import has_focus
from prompt_toolkit.formatted_text import FormattedText, HTML, to_formatted_text
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout.containers import (
    DynamicContainer,
    HorizontalAlign,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.widgets import Box, Button, Frame, Label, Shadow


class WizardDialog:
    def __init__(  # noqa: CCR001
        self, title, handlers, intro=None, summary=False,
        next_text='Next', previous_text='Previous',
        cancel_text='Cancel', finish_text='Finish',
    ):
        self.title = title
        self.handlers = handlers
        self.intro = intro
        self.summary = summary
        self.steps = []
        self.current_step_idx = 0
        self.title = title
        self.process_steps()
        self.current_step = self.steps[self.current_step_idx]
        self.label_next = next_text
        self.label_previous = previous_text
        self.label_cancel = cancel_text
        self.label_finish = finish_text
        self.error_messages = ''

        self.cancel_btn = Button(
            text=self.label_cancel,
            handler=self.cancel,
        )
        self.previous_btn = Button(
            text=self.label_previous,
            handler=self.previous,
        )
        self.next_btn = Button(
            text=self.label_next if len(self.steps) > 1 else self.label_finish,
            handler=self.next,
        )

        self.buttons = [self.next_btn, self.cancel_btn]

        self.buttons_kb = KeyBindings()
        first_selected = has_focus(self.buttons[0])
        last_selected = has_focus(self.buttons[-1])

        self.buttons_kb.add('left', filter=~first_selected)(focus_previous)
        self.buttons_kb.add('right', filter=~last_selected)(focus_next)

        input_container = HSplit(
            [
                Box(
                    body=DynamicContainer(self.get_current_step_container),
                    padding=D(preferred=1, max=1),
                    padding_bottom=0,
                ),
            ],
        )

        left_container = Box(
            body=DynamicContainer(self.get_steps_labels),
            padding=D(preferred=1, max=1),
            padding_bottom=0,
        )
        right_container = HSplit(
            [
                input_container,
                Box(
                    body=DynamicContainer(self.get_status),
                    padding=D(preferred=1, max=1),
                    padding_bottom=1,
                ),
            ],
        )
        top_container = VSplit(
            [left_container, right_container],
            padding_char='│',
            padding=1,
            padding_style='#000000',
            height=D(min=10, preferred=24),
        )

        buttons_container = Box(
            body=DynamicContainer(self.get_buttons_container),
            height=D(min=1, max=3, preferred=3),
        )

        kb = KeyBindings()
        kb.add(Keys.Tab)(focus_next)
        kb.add(Keys.BackTab)(focus_previous)

        frame = Shadow(
            body=Frame(
                title=self.get_title,
                body=HSplit(
                    [
                        top_container,
                        buttons_container,
                    ],
                    padding_char='─',
                    padding=1,
                    padding_style='#000000',
                ),
                style='class:dialog.body',
                key_bindings=kb,
                width=D(min=78, preferred=132),
            ),
        )
        self.container = Box(
            body=frame, style='class:dialog',
        )

    def get_title(self):
        return (
            f'{self.title} - {self.current_step_idx + 1} of {len(self.steps)}'
        )

    def get_steps_labels(self):
        steps_labels = []
        for idx, step in enumerate(self.steps, start=1):
            label = f'{idx}. {step["label"]}'
            if self.current_step_idx == idx - 1:
                label = f'<magenta><b>{label}</b></magenta>'
            steps_labels.append(
                Window(
                    FormattedTextControl(
                        to_formatted_text(HTML(label)),
                    ),
                    height=1,
                ),
            )
        return HSplit(steps_labels, width=20)

    def get_status(self):
        if self.error_messages:
            return Window(
                FormattedTextControl(
                    FormattedText([('class:error', self.error_messages)]),
                ),
                height=1,
            )
        return Label('')

    def get_summary(self):
        text = '\n'.join(
            [
                f'<b>{handler.get_variable_name().capitalize()}:'
                f' </b>{handler.get_formatted_value()}'
                for handler in self.handlers
            ],
        )
        return Window(
            FormattedTextControl(to_formatted_text(HTML(text))),
            wrap_lines=True,
        )

    def get_current_step_container(self):
        return self.current_step['layout']

    def get_buttons_container(self):
        return VSplit(
            self.buttons,
            padding=1,
            key_bindings=self.buttons_kb,
        )

    def process_steps(self):
        if self.intro:
            layout = Window(
                FormattedTextControl(to_formatted_text(HTML(self.intro))),
                wrap_lines=True,
            )
            self.steps.append(
                {
                    'layout': layout,
                    'label': 'Introduction',
                    'handler': None,
                },
            )

        for handler in self.handlers:
            layout = handler.get_layout()
            layout.align = HorizontalAlign.JUSTIFY
            self.steps.append(
                {
                    'layout': layout,
                    'label': handler.get_variable_name().capitalize(),
                    'handler': handler,
                },
            )

        if self.summary:
            layout = Box(
                body=DynamicContainer(self.get_summary),
                padding=D(preferred=1, max=1),
                padding_bottom=1,
            )
            self.steps.append(
                {
                    'layout': layout,
                    'label': 'Summary',
                    'handler': None,
                },
            )

    def set_buttons_labels(self):
        if len(self.steps) == 1:
            return
        if self.current_step_idx == 0:
            self.buttons = [self.next_btn, self.cancel_btn]
            return
        if self.current_step_idx == len(self.steps) - 1:
            self.next_btn.text = self.label_finish
        else:
            self.next_btn.text = self.label_next

        self.buttons = [self.next_btn, self.previous_btn, self.cancel_btn]

    def cancel(self):
        get_app().exit(result=False)

    def previous(self):
        if self.current_step_idx != 0:
            self.error_messages = ''
            self.current_step_idx -= 1
            self.current_step = self.steps[self.current_step_idx]
            get_app().layout.focus(self.current_step['layout'])

        self.set_buttons_labels()

    def next(self):
        if self.validate():
            if self.current_step_idx < len(self.steps) - 1:
                self.current_step_idx += 1
                self.current_step = self.steps[self.current_step_idx]
                if not self.summary or self.current_step != self.steps[-1]:
                    get_app().layout.focus(self.current_step['layout'])
            else:
                get_app().exit(result=True)

        self.set_buttons_labels()

    def validate(self):
        step = self.steps[self.current_step_idx]
        handler = step['handler']
        if handler and not handler.is_valid():
            self.error_messages = ','.join(handler.errors)

            return False
        self.error_messages = ''
        return True

    def __pt_container__(self):  # pragma: no cover
        return self.container



# pylint: disable=unused-argument
import string

from prompt_toolkit.application.current import get_app
from prompt_toolkit.keys import Keys
from prompt_toolkit.layout import HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.widgets import Box, Button, Label, TextArea

from interrogatio.handlers.base import QHandler
from interrogatio.handlers.registry import register
from interrogatio.widgets import DateRange, MaskedInput, SelectMany, SelectOne


@register('input')
class StringHandler(QHandler):

    def get_widget_class(self):
        return TextArea

    def get_widget_init_kwargs(self):
        kwargs = dict(
            multiline=self._question.get('multiline', False),
            style='class:input.answer',
        )
        if 'default' in self._question:
            kwargs['text'] = self._question['default']
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        widget.buffer.cursor_position = len(widget.text)
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_value(self):
        return self.get_widget().text


@register('password')
class PasswordHandler(QHandler):

    def get_widget_class(self):
        return TextArea

    def get_widget_init_kwargs(self):
        kwargs = dict(
            multiline=False,
            style='class:password.answer',
            password=True,
        )
        if 'default' in self._question:
            kwargs['text'] = self._question['default']
        return kwargs

    def get_value(self):
        return self.get_widget().text

    def get_layout(self):
        widget = self.get_widget()
        widget.buffer.cursor_position = len(widget.text)

        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:password.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:password.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)


@register('selectone')
class SelectOneHandler(QHandler):

    def get_widget_class(self):
        return SelectOne

    def get_value(self):
        return self.get_widget().current_value

    def get_widget_init_kwargs(self):
        kwargs = dict(
            values=self._question['values'],
            style='class:selectone.answer',
        )
        if 'default' in self._question:
            kwargs['default'] = self._question['default']

        return kwargs

    def get_layout(self):
        vsplit_components = [self.get_widget()]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    dont_extend_height=False,
                    style='class:selectone.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:selectone.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_formatted_value(self):
        value = self.get_value()
        options = {t[0]: t[1] for t in self._question['values']}
        return f'{options[value]} ({value})'


@register('selectmany')
class SelectManyHandler(QHandler):

    def get_widget_class(self):
        return SelectMany

    def get_value(self):
        return list(self.get_widget().checked)

    def get_widget_init_kwargs(self):
        kwargs = dict(
            values=self._question['values'],
            style='class:selectmany.answer',
        )
        if 'checked' in self._question:
            kwargs['checked'] = set(self._question['checked'])
        if 'default' in self._question:
            kwargs['default'] = self._question['default']

        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    dont_extend_height=False,
                    style='class:selectone.question',
                ),
            )

        btn_all = Button('All', widget.select_all)
        btn_none = Button('None', widget.select_none)
        buttons = Box(
            body=VSplit([btn_all, btn_none], padding=1),
            height=D(min=1, max=3, preferred=3),
            padding_left=0,
        )

        hsplit_components = [buttons, VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:selectone.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_keybindings(self):

        bindings = super().get_keybindings()

        @bindings.add(Keys.Tab)
        def _tab(event):
            get_app().layout.focus_next()

        @bindings.add(Keys.BackTab)
        def _s_tab(event):
            get_app().layout.focus_previous()

        return bindings

    def get_formatted_value(self):
        values = self.get_value()
        if not values:
            return ''
        options = {t[0]: t[1] for t in self._question['values']}

        formatted_values = []

        for idx, value in enumerate(values):
            if idx > 14:
                formatted_values.append('...')
                break
            formatted_values.append(f'{options[value]} ({value})')

        return ', '.join(formatted_values)


@register('maskedinput')
class MaskedInputHandler(QHandler):

    def get_widget_class(self):
        return MaskedInput

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
            mask=self._question['mask'],
        )
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_value(self):
        return self.get_widget().value


@register('date')
class DateHandler(QHandler):

    def get_widget_class(self):
        return MaskedInput

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
            mask='____-__-__',
            allowed_chars=string.digits,
        )
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_value(self):
        return self.get_widget().value


@register('daterange')
class DateRangeHandler(QHandler):

    def get_widget_class(self):
        return DateRange

    def get_widget_init_kwargs(self):
        kwargs = dict(
            style='class:input.answer',
        )
        if 'from_label' in self._question:
            kwargs['from_label'] = self._question['from_label']
        if 'to_label' in self._question:
            kwargs['to_label'] = self._question['to_label']
        return kwargs

    def get_layout(self):
        widget = self.get_widget()
        vsplit_components = [widget]
        if 'message' in self._question and self._question['message']:
            vsplit_components.insert(
                0,
                Label(
                    self._question['message'],
                    dont_extend_width=True,
                    dont_extend_height=False,
                    style='class:input.question',
                ),
            )
        hsplit_components = [VSplit(vsplit_components, padding=1)]
        if 'description' in self._question and self._question['description']:
            hsplit_components.insert(
                0,
                Label(
                    self._question['description'],
                    style='class:input.question',
                ),
            )
        return HSplit(hsplit_components, padding=1)

    def get_value(self):
        return self.get_widget().value

    def get_formatted_value(self):
        value = self.get_value()
        return f'{value["from"]} - {value["to"]}'

    def get_keybindings(self):

        bindings = super().get_keybindings()

        @bindings.add(Keys.Tab)
        def _tab(event):
            get_app().layout.focus_next()

        @bindings.add(Keys.BackTab)
        def _s_tab(event):
            get_app().layout.focus_previous()

        return bindings



from prompt_toolkit.shortcuts import button_dialog as pt_button_dialog
from prompt_toolkit.shortcuts import input_dialog as pt_input_dialog
from prompt_toolkit.shortcuts import message_dialog as pt_message_dialog
from prompt_toolkit.shortcuts import progress_dialog as pt_progress_dialog
from prompt_toolkit.shortcuts import radiolist_dialog as pt_radiolist_dialog
from prompt_toolkit.shortcuts import yes_no_dialog as pt_yes_no_dialog

from interrogatio.themes import for_dialog

__all__ = [
    'yes_no_dialog',
    'button_dialog',
    'input_dialog',
    'message_dialog',
    'radiolist_dialog',
    'progress_dialog',
]


def yes_no_dialog(
    title='',
    text='',
    yes_text='Yes',
    no_text='No',
    style=None,
):
    style = style or for_dialog()
    return pt_yes_no_dialog(
        title=title,
        text=text,
        yes_text=yes_text,
        no_text=no_text,
        style=style,
    )


def button_dialog(
    title='',
    text='',
    buttons=None,
    style=None,
):
    style = style or for_dialog()
    return pt_button_dialog(
        title=title,
        text=text,
        buttons=buttons or [],
        style=style,
    )


def input_dialog(
    title='', text='', ok_text='OK', cancel_text='Cancel',
    completer=None, password=False, style=None,
):
    style = style or for_dialog()
    return pt_input_dialog(
        title=title, text=text, ok_text=ok_text,
        cancel_text=cancel_text, completer=completer,
        password=password, style=style,
    )


def message_dialog(
    title='', text='', ok_text='Ok', style=None,
):
    style = style or for_dialog()
    return pt_message_dialog(
        title=title, text=text,
        ok_text=ok_text, style=style,
    )


def radiolist_dialog(
    title='', text='', ok_text='Ok', cancel_text='Cancel',
    values=None, style=None,
):
    style = style or for_dialog()
    return pt_radiolist_dialog(
        title=title, text=text,
        ok_text=ok_text, cancel_text=cancel_text,
        values=values, style=style,
    )


def progress_dialog(
    title='', text='',
    run_callback=None, style=None,
):
    style = style or for_dialog()
    return pt_progress_dialog(
        title=title, text=text,
        run_callback=run_callback, style=style,
    )



from prompt_toolkit.application import Application
from prompt_toolkit.layout import Layout

from interrogatio.core.utils import validate_questions
from interrogatio.handlers import get_instance
from interrogatio.themes import for_dialog, set_theme
from interrogatio.widgets.wizard import WizardDialog


__all__ = ['dialogus']


def show_dialog(
    questions,
    title,
    intro=None,
    summary=False,
    next_text='Next',
    previous_text='Previous',
    cancel_text='Cancel',
    finish_text='Finish',
):
    handlers = [get_instance(q) for q in questions]
    app = Application(
        layout=Layout(
            WizardDialog(
                title, handlers,
                intro=intro, summary=summary,
                next_text=next_text, previous_text=previous_text,
                cancel_text=cancel_text, finish_text=finish_text,
            ),
        ),
        mouse_support=True,
        style=for_dialog(),
        full_screen=True,
    )

    if not app.run():
        return
    answers = {}
    for handler in handlers:
        answers.update(handler.get_answer())
    return answers


def dialogus(
    questions,
    title,
    intro=None,
    summary=False,
    next_text='Next',
    previous_text='Previous',
    cancel_text='Cancel',
    finish_text='Finish',
    theme='default',
):
    """
    Show a dialog with inputs as defined in the questions parameter and returns
    a dictionary with the answers.

    :param questions: a list of questions.
    :type questions: list

    :param title: the title of the dialog.
    :type title: str

    :param confirm: the confirm button text.
    :type confirm: str

    :param cancel: the cancel button text.
    :type cancel: str

    :return: a dictionary with the answers.
    :rtype: dict

    Usage:

    .. code-block:: python

        from interrogatio import dialogus
        questions = [
            {
                'name': 'name',
                'type': 'input',
                'message': 'What is your name'
            },
            {
                'name': 'favorite_pet',
                'type': 'input',
                'message': 'What is your favorite pet'
            }
        ]
        answers = dialogus(
            questions,
            title='Tell me something about you',
            finish_text='Done',
            cancel_text='Cancel',
        )
    """
    set_theme(theme)
    validate_questions(questions)
    return show_dialog(
        questions,
        title,
        intro=intro,
        summary=summary,
        next_text=next_text,
        previous_text=previous_text,
        cancel_text=cancel_text,
        finish_text=finish_text,
    )



import argparse
import json
import sys
from functools import partial

try:
    import yaml
    FORMAT_CHOICES = ['json', 'yaml']
except ImportError:
    yaml = None
    FORMAT_CHOICES = ['json']


from interrogatio import dialogus, interrogatio


def _load_questions(args):
    with args.input as f:
        return args.deserialize(f)


def _write_answers(args, answers):
    with args.output as f:
        args.serialize(answers, f)


def _add_common_arguments(parser):
    parser.add_argument(
        '--input',
        '-i',
        type=argparse.FileType('r'),
        required=True,
        help='Input file with questions',
    )
    parser.add_argument(
        '--output',
        '-o',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output file to write answers to (Default: STDOUT)',
    )
    if len(FORMAT_CHOICES) > 1:
        parser.add_argument(
            '--input-format',
            choices=FORMAT_CHOICES,
            default='json',
            help='Questions file format (Default: json)',
        )
        parser.add_argument(
            '--output-format',
            choices=FORMAT_CHOICES,
            default='json',
            help='Answers file format (Default: json)',
        )
    parser.add_argument(
        '--theme',
        '-t',
        default='default',
        help='Name of the UI theme to use (Default: default)',
    )


def main_dialogus():
    parser = argparse.ArgumentParser(
        description='Show a wizard dialog to prompt user for questions.',
    )

    _add_common_arguments(parser)

    parser.add_argument(
        '--title',
        help='Title of the dialog',
    )
    parser.add_argument(
        '--intro',
        help='Specify the text of the introduction step (Default: no intro)',
    )
    parser.add_argument(
        '--summary',
        action='store_true',
        help=(
            'Show a summary with answers as the '
            'latest step (Default: no summary)'
        ),
    )

    for button in ('previous', 'next', 'cancel', 'finish'):
        cap_btn = button.capitalize()
        parser.add_argument(
            f'--{button}',
            default=cap_btn,
            help=(
                f'Customize the text of the "{button}" '
                f'button (Default: {cap_btn})'
            ),
        )

    args = parser.parse_args()

    if args.input_format == 'yaml':
        args.deserialize = partial(yaml.load, Loader=yaml.FullLoader)
        args.serialize = yaml.dump
    else:
        args.deserialize = json.load
        args.serialize = json.dump

    kwargs = {
        'intro': args.intro,
        'summary': args.summary,
        'title': args.title,
        'previous_text': args.previous,
        'next_text': args.next,
        'cancel_text': args.cancel,
        'finish_text': args.finish,
    }

    _write_answers(args, dialogus(_load_questions(args), **kwargs))


def main_interrogatio():
    parser = argparse.ArgumentParser(
        description='Prompt user for questions.',
    )
    _add_common_arguments(parser)

    args = parser.parse_args()

    if args.input_format == 'yaml':
        args.deserialize = partial(yaml.load, Loader=yaml.FullLoader)
        args.serialize = yaml.dump
    else:
        args.deserialize = json.load
        args.serialize = json.dump

    _write_answers(
        args,
        interrogatio(_load_questions(args), theme=args.theme),
    )



from interrogatio.handlers.base import QHandler  # noqa
from interrogatio.handlers.registry import (  # noqa
    register, get_instance, get_registered,
)
from interrogatio.handlers.builtins import *  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['interrogatio',
 'interrogatio.core',
 'interrogatio.handlers',
 'interrogatio.shortcuts',
 'interrogatio.themes',
 'interrogatio.validators',
 'interrogatio.widgets']

package_data = \
{'': ['*'], 'interrogatio.themes': ['theme_files/*']}

install_requires = \
['prompt-toolkit>=3.0.0,<4.0.0', 'validators>=0.18.2,<0.19.0']

extras_require = \
{'yaml': ['PyYAML>=5']}

entry_points = \
{'console_scripts': ['dialogus = interrogatio.main:main_dialogus',
                     'interrogatio = interrogatio.main:main_interrogatio']}

setup_kwargs = {
    'name': 'interrogatio',
    'version': '2.0.0',
    'description': 'Prompting library for terminals.',
    'long_description': '#\xa0interrogatio\n\n![Python versions](https://img.shields.io/pypi/pyversions/interrogatio.svg) [![PyPi Status](https://img.shields.io/pypi/v/interrogatio.svg)](https://pypi.org/project/interrogatio/) ![Read the Docs](https://img.shields.io/readthedocs/interrogatio) [![Build Status](https://img.shields.io/github/workflow/status/ffaraone/interrogatio/Build%20interrogatio)](https://github.com/ffaraone/interrogatio/actions) [![codecov](https://codecov.io/gh/ffaraone/interrogatio/branch/master/graph/badge.svg)](https://codecov.io/gh/ffaraone/interrogatio)\n\n\nA python library to prompt users for inputs in a terminal application.\n\n\n## What is interrogatio\n\n`interrogatio` is a python 3.8+ library based on the [python-prompt-toolkit](https://github.com/prompt-toolkit/python-prompt-toolkit) and inspired by [PyInquirer](https://github.com/CITGuru/PyInquirer/) that help CLI developers to ask users for inputs.\n\nQuestions can be rendered onto the terminal prompt or as curses-like dialogs.\n\n\n## Documentation\n\n[`interrogatio` documentation](https://interrogatio.readthedocs.io/en/latest/) is hosted on _Read the Docs_.\n\n\n\n## Getting started\n\n\n### Requirements\n\n`interrogatio` depends on the python-prompt-toolkit library and its dependencies.\n\n### Installation\n\n\n#### Using pip\n\n\n```\n$ pip install interrogatio\n```\n\n\n#### Extra dependencies\n\nIf you want to use the shell command with yml files you can install the yml dependency:\n\n```\n$ pip install interrogatio[yml]\n```\n\n\n### Basic usage\n\n`interrogatio` needs a list of questions to prompt the user for answers.\n\nEach question is a python dictionary with at least the following keys:\n\n* **name**: it has to be unique within the list of questions. It represents the variable name;\n* **type**: the type of question;\n* **message**: the text of the prompt.\n\nOptionally you should specify:\n    \n* a **default**: a default value;\n* a **validators**: a list of children of Validator class\n* a **values**: a list of tuples (value, label) to provide a list of choices \n    for the ``selectone`` or ``selectmany`` question types.\n\n\n`interrogatio` can run into two modes: dialog and prompt.\n\n#### Dialog mode\n\n![Dialog mode showcase](docs/showcase/dialogus.gif)\n\n```\nfrom interrogatio import dialogus\n\nquestions = [\n    {\n        \'name\': \'name\',\n        \'type\': \'input\',\n        \'message\': "What\'s your name ?",\n        \'description\': \'Please enter your full name. This field is required.\',\n        \'validators\': [{\'name\': \'required\'}],\n    },\n    {\n        \'name\': \'birth_date\',\n        \'type\': \'date\',\n        \'message\': "What\'s your birth date ?",\n        \'description\': \'Enter your birth date.\',\n    },\n    {\n        \'name\': \'nationality\',\n        \'type\': \'selectone\',\n        \'message\': "What\'s your nationality ?",\n        \'description\': \'Please choose one from the list.\',\n        \'validators\': [{\'name\': \'required\'}],\n        \'values\': [\n            (\'IT\', \'Italian\'),\n            (\'ES\', \'Spanish\'),\n            (\'US\', \'American\'),\n            (\'UK\', \'English\'),\n        ],\n    },\n    {\n        \'name\': \'languages\',\n        \'type\': \'selectmany\',\n        \'message\': "What are your favorite programming languages ?",\n        \'description\': \'Please choose your favorites from the list.\',\n        \'values\': [\n            (\'py\', \'Python\'),\n            (\'rb\', \'Ruby\'),\n            (\'js\', \'Javascript\'),\n            (\'go\', \'Golang\'),\n            (\'rs\', \'Rust\'),\n            (\'c\', \'C\'),\n            (\'cpp\', \'C++\'),\n            (\'java\', \'Java\'),\n        ],\n    },\n]\n\nintro = """<blue>Welcome to <b><i>interrogatio 2.0</i></b>!\n\nThis is the second major release of interrogatio with nice improvements.</blue>\n\n<b>What\'s new</b>\n<b>----------</b>\n\n* Curses-like dialog experience had been completely rewritten.\n* New questions handlers for dates, date ranges and masked inputs.\n* Validators are now based on the <u>validators</u> library.\n"""\n\n\nanswers = dialogus(questions, \'interrogatio showcase\', intro=intro, summary=True)\n```\n\n#### Prompt mode\n\n![Prompt mode showcase](docs/showcase/interrogatio.gif)\n\n```\nfrom interrogatio import interrogatio\n\nquestions = [\n    {\n        \'name\': \'name\',\n        \'type\': \'input\',\n        \'message\': "What\'s your name ?",\n        \'description\': \'Please enter your full name. This field is required.\',\n        \'validators\': [{\'name\': \'required\'}],\n    },\n    {\n        \'name\': \'birth_date\',\n        \'type\': \'date\',\n        \'message\': "What\'s your birth date ?",\n        \'description\': \'Enter your birth date.\',\n    },\n    {\n        \'name\': \'nationality\',\n        \'type\': \'selectone\',\n        \'message\': "What\'s your nationality ?",\n        \'description\': \'Please choose one from the list.\',\n        \'validators\': [{\'name\': \'required\'}],\n        \'values\': [\n            (\'IT\', \'Italian\'),\n            (\'ES\', \'Spanish\'),\n            (\'US\', \'American\'),\n            (\'UK\', \'English\'),\n        ],\n    },\n    {\n        \'name\': \'languages\',\n        \'type\': \'selectmany\',\n        \'message\': "What are your favorite programming languages ?",\n        \'description\': \'Please choose your favorites from the list.\',\n        \'values\': [\n            (\'py\', \'Python\'),\n            (\'rb\', \'Ruby\'),\n            (\'js\', \'Javascript\'),\n            (\'go\', \'Golang\'),\n            (\'rs\', \'Rust\'),\n            (\'c\', \'C\'),\n            (\'cpp\', \'C++\'),\n            (\'java\', \'Java\'),\n        ],\n    },\n]\n\n\nanswers = interrogatio(questions)\n```\n\n### Contributing\n\nIf you want to contribute to the project, you can submit bugs, feature requests or fork the github repository and submit your pull request.\n\n\n### License\n\n`interrogatio` is released under the [BSD 3-Clause "New" or "Revised" License](https://opensource.org/licenses/BSD-3-Clause>).\n',
    'author': 'Francesco Faraone',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ffaraone/interrogatio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from datetime import datetime

import validators

from interrogatio.core.exceptions import ValidationError
from interrogatio.validators.base import Validator
from interrogatio.validators.registry import register


__all__ = [
    'RequiredValidator',
    'RegexValidator',
    'URLValidator',
    'EmailValidator',
    'MinLengthValidator',
    'MaxLengthValidator',
    'NumberValidator',
    'IntegerValidator',
    'IPv4Validator',
    'RangeValidator',
    'MinValidator',
    'MaxValidator',
    'DateTimeValidator',
    'DateTimeRangeValidator',
]


@register('required')
class RequiredValidator(Validator):
    def __init__(self, message=None):
        """
        Initialise the ``required`` validator.

        :param message: the error message in case that validation fails.
        :type message: str
        """
        super(RequiredValidator, self).__init__(
            message=message or 'this field is required',
        )

    def validate(self, value):
        if not value:
            raise ValidationError(self.message)


@register('regex')
class RegexValidator(Validator):
    def __init__(self, expr, message=None, inverse_match=None):
        """
        Initialise the ``regex`` validator.

        :param expr: the regex to match.
        :type expr: str
        :param message: the error message in case that validation fails.
        :type message: str
        :param inverse_match: invert the match of the expression.
        :type inverse_match: bool
        """
        super(RegexValidator, self).__init__(
            message=message or f'this field does not match {expr}',
        )
        self.regex = re.compile(expr)
        self.inverse_match = inverse_match

    def validate(self, value):
        regex_matches = self.regex.search(str(value))
        invalid_input = (
            regex_matches if self.inverse_match else not regex_matches
        )
        if invalid_input:
            raise ValidationError(self.message)


@register('email')
class EmailValidator(Validator):

    def __init__(self, message=None):
        """
        Initialise the ``email`` validator.
        :param message: the error message in case that validation fails.
        :type message: str
        """
        super(EmailValidator, self).__init__(
            message=message or 'this field must be an email address',
        )

    def validate(self, value):
        if value and validators.email(value) is not True:
            raise ValidationError(self.message)


@register('url')
class URLValidator(Validator):

    def __init__(self, message=None):
        super().__init__(message=message or 'this field must be an url')

    def validate(self, value):
        if value and validators.url(value) is not True:
            raise ValidationError(self.message)


@register('min-length')
class MinLengthValidator(Validator):
    def __init__(self, min_length, message=None):
        """
        Initialise the ``min-length`` validator.

        :param min_length: Minumum length from which the value has to be
                           considered valid.
        :type min_length: int
        :param message: the error message in case that validation fails.
        :type message: str
        """

        self.min_length = min_length
        self.message = message or (
            'the length of this field must be at '
            'least {} characters long'.format(self.min_length)
        )

    def validate(self, value):
        try:
            if (
                value
                and validators.length(value, min=self.min_length) is not True
            ):
                raise ValidationError(message=self.message)
        except (TypeError, AssertionError):
            raise ValidationError(message=self.message)


@register('max-length')
class MaxLengthValidator(Validator):
    def __init__(self, max_length, message=None):
        """
        Initialise the ``max-length`` validator.

        :param max_length: Maximum length from which the value has to be
                           considered invalid.
        :type max_length: int
        :param message: the error message in case that validation fails.
        :type message: str
        """
        self.max_length = max_length
        self.message = message or (
            'the length of this field must be at '
            f'most {max_length} characters long'
        )

    def validate(self, value):
        try:
            if (
                value
                and validators.length(value, max=self.max_length) is not True
            ):
                raise ValidationError(message=self.message)
        except (TypeError, AssertionError):
            raise ValidationError(message=self.message)


@register('number')
class NumberValidator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be a number'

    def validate(self, value):
        try:
            float(value)
        except (ValueError, TypeError):
            raise ValidationError(message=self.message)


@register('integer')
class IntegerValidator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be an integer'

    def validate(self, value):
        try:
            int(value)
        except (ValueError, TypeError):
            raise ValidationError(message=self.message)

        if int(value) != value:
            raise ValidationError(message=self.message)


@register('ipv4')
class IPv4Validator(Validator):
    def __init__(self, message=None):
        self.message = message or 'this field must be an IPv4 address'

    def validate(self, value):
        if value and validators.ipv4(value) is not True:
            raise ValidationError(message=self.message)


@register('range')
class RangeValidator(Validator):
    def __init__(self, min=None, max=None, message=None):
        self.min = min
        self.max = max
        self.message = message or (
            f'this field must be a number between {min} and {max}'
        )

    def validate(self, value):
        try:
            if (
                value is not None and value != ''
                and validators.between(
                    value,
                    min=self.min,
                    max=self.max,
                ) is not True
            ):
                raise ValidationError(message=self.message)
        except (TypeError, AssertionError):
            raise ValidationError(message=self.message)


@register('min')
class MinValidator(Validator):
    def __init__(self, min=None, message=None):
        self.min = min
        self.message = (
            message or f'this field must be greater or equal to {min}'
        )

    def validate(self, value):
        try:
            if (
                value is not None and value != ''
                and validators.between(
                    value,
                    min=self.min,
                ) is not True
            ):
                raise ValidationError(message=self.message)
        except (TypeError, AssertionError):
            raise ValidationError(message=self.message)


@register('max')
class MaxValidator(Validator):
    def __init__(self, max=None, message=None):
        self.max = max
        self.message = (
            message or f'this field must be smaller or equal to {max}'
        )

    def validate(self, value):
        try:
            if (
                value is not None and value != ''
                and validators.between(
                    value,
                    max=self.max,
                ) is not True
            ):
                raise ValidationError(message=self.message)
        except (TypeError, AssertionError):
            raise ValidationError(message=self.message)


@register('datetime')
class DateTimeValidator(Validator):
    def __init__(self, format_pattern='%Y-%m-%dT%H:%M:%S', message=None):
        self.format_pattern = format_pattern
        self.message = message or 'this field is not a valid datetime'

    def validate(self, value):
        if value is None or value == '':
            return
        try:
            datetime.strptime(value, self.format_pattern)
        except (TypeError, ValueError):
            raise ValidationError(message=self.message)


@register('datetimerange')
class DateTimeRangeValidator(Validator):
    def __init__(self, format_pattern='%Y-%m-%d', message=None):
        self.format_pattern = format_pattern
        self.message = message or 'this field is not a valid datetime range'

    def validate(self, value):
        if value is None or value == '':
            return
        d_from = None
        d_to = None
        try:
            if value['from']:
                d_from = datetime.strptime(value['from'], self.format_pattern)
            if value['to']:
                d_to = datetime.strptime(value['to'], self.format_pattern)
        except (TypeError, ValueError):
            raise ValidationError(message=self.message)

        if d_from and d_to:
            if d_from > d_to:
                raise ValidationError(message=self.message)
